@@ -43,12 +43,18 @@
 
 #include <QtCore/qatomic.h>
 #include <QtCore/qthread.h>
+#include <QtCore/private/qsimd_p.h> // for qYieldCpu()
 #include <private/qthreadpool_p.h>
 
 #ifdef interface
 #  undef interface
 #endif
 
+// GCC 12 gets confused about QFutureInterfaceBase::state, for some non-obvious
+// reason
+//  warning: ‘unsigned int __atomic_or_fetch_4(volatile void*, unsigned int, int)’ writing 4 bytes into a region of size 0 overflows the destination [-Wstringop-overflow=]
+QT_WARNING_DISABLE_GCC("-Wstringop-overflow")
+
 QT_BEGIN_NAMESPACE
 
 enum {
@@ -102,12 +108,11 @@ static inline int switch_off(QAtomicInt &a, int which)
 
 static inline int switch_from_to(QAtomicInt &a, int from, int to)
 {
-    int newValue;
-    int expected = a.loadRelaxed();
-    do {
-        newValue = (expected & ~from) | to;
-    } while (!a.testAndSetRelaxed(expected, newValue, expected));
-    return newValue;
+    const auto adjusted = [&](int old) { return (old & ~from) | to; };
+    int value = a.loadRelaxed();
+    while (!a.testAndSetRelaxed(value, adjusted(value), value))
+        qYieldCpu();
+    return value;
 }
 
 void QFutureInterfaceBase::cancel()
